# -*- coding: utf8 -*-
import sys
import os.path
import numpy as np
import PDielec.Calculator as Calculator
from PyQt5.QtWidgets  import  QPushButton, QWidget
from PyQt5.QtWidgets  import  QComboBox, QLabel, QLineEdit, QDoubleSpinBox
from PyQt5.QtWidgets  import  QVBoxLayout, QHBoxLayout, QFormLayout
from PyQt5.QtWidgets  import  QSpinBox
from PyQt5.QtCore     import  Qt
from PDielec.Constants import  support_matrix_db, wavenumber, amu, PI
from PDielec.Constants import  avogadro_si
from PDielec.Utilities import  Debug

class ScenarioTab(QWidget):
    def __init__(self, parent, debug=False):
        super(QWidget, self).__init__(parent)
        global debugger
        debugger = Debug(debug,'ScenarioTab:')
        self.dirty = True
        self.settings = {}
        self.notebook = parent
        self.notebook.plottingCalculationRequired = True
        self.notebook.fittingCalculationRequired = True
        matrix = 'ptfe'
        self.settings['Matrix'] = matrix
        self.settings['Matrix density'] = support_matrix_db[matrix][0]
        self.settings['Matrix permittivity'] = support_matrix_db[matrix][1]
        self.settings['Bubble radius'] = 30.0
        self.settings['Bubble volume fraction'] = 0.0
        self.settings['Mass fraction'] = 0.1
        self.settings['Volume fraction'] = 0.1
        self.settings['Particle size(mu)'] = 0.0001
        self.settings['Particle size distribution sigma(mu)'] = 0.0
        self.settings['Ellipsoid a/b'] = 1.0
        self.settings['Unique direction - h'] = 0
        self.settings['Unique direction - k'] = 0
        self.settings['Unique direction - l'] = 1
        self.settings['Mass or volume fraction'] = 'volume'
        self.settings['ATR material refractive index'] = 4.0
        self.settings['ATR theta'] = 45.0
        self.settings['ATR S polarisation fraction'] = 0.5
        # get the reader from the main tab
        self.notebook = parent
        self.reader = self.notebook.mainTab.reader
        self.settings['Effective medium method'] = 'Maxwell-Garnett'
        # self.methods = ['Maxwell-Garnett', 'Bruggeman', 'Averaged Permittivity', 'Mie', 'Anisotropic-Mie']
        self.methods = ['Maxwell-Garnett', 'Bruggeman', 'Averaged Permittivity', 'Mie']
        self.settings['Particle shape'] = 'Sphere'
        self.shapes = ['Sphere', 'Needle', 'Plate', 'Ellipsoid']
        self.scenarioIndex = None
        # Create a scenario tab
        vbox = QVBoxLayout()
        form = QFormLayout()
        #
        # Support matrix
        #
        self.matrix_cb = QComboBox(self)
        self.matrix_cb.setToolTip('Define the permittivity and density of the support matrix')
        self.matrix_cb.addItems(support_matrix_db)
        index = self.matrix_cb.findText(self.settings['Matrix'], Qt.MatchFixedString)
        if index >=0:
            self.matrix_cb.setCurrentIndex(index)
        else:
            print('support matrix index was not 0',matrix)
        self.matrix_cb.activated.connect(self.on_matrix_cb_activated)
        label = QLabel('Support matrix',self)
        label.setToolTip('Define the permittivity and density of the support matrix')
        form.addRow(label, self.matrix_cb)
        #
        # Support matrix permittivity
        #
        self.density_sb = QDoubleSpinBox(self)
        self.density_sb.setRange(0.001, 100.0)
        self.density_sb.setSingleStep(0.01)
        self.density_sb.setDecimals(3)
        self.density_sb.setToolTip('Define the support matrix density. \nThis makes changes to the support density and permittivity')
        self.density_sb.setValue(self.settings['Matrix density'])
        self.density_sb.valueChanged.connect(self.on_density_sb_changed)
        label = QLabel('Support density', self)
        label.setToolTip('Define the support matrix density. \nThis makes changes to the support density and permittivity')
        form.addRow(label, self.density_sb)
        #
        # Support matrix permittivity
        #
        self.permittivity_sb = QDoubleSpinBox(self)
        self.permittivity_sb.setRange(0.001, 100.0)
        self.permittivity_sb.setSingleStep(0.01)
        self.permittivity_sb.setDecimals(3)
        self.permittivity_sb.setToolTip('Define the support matrix permittivity')
        self.permittivity_sb.setValue(self.settings['Matrix permittivity'])
        self.permittivity_sb.valueChanged.connect(self.on_permittivity_sb_changed)
        label = QLabel('Support permittivity', self)
        label.setToolTip('Define the support matrix permittivity')
        form.addRow(label, self.permittivity_sb)
        #
        # Bubble volume fraction
        #
        self.bubble_vf_sb = QDoubleSpinBox(self)
        self.bubble_vf_sb.setRange(0.0, 100.0*(1.0-self.settings['Volume fraction']))
        self.bubble_vf_sb.setSingleStep(1.0)
        self.bubble_vf_sb.setDecimals(1)
        self.bubble_vf_sb.setToolTip('Define the % volume fraction of air bubble inclusions in the matrix')
        self.bubble_vf_sb.setValue(100*self.settings['Bubble volume fraction'])
        self.bubble_vf_sb.valueChanged.connect(self.on_bubble_vf_sb_changed)
        label = QLabel('% Air void volume fraction', self)
        label.setToolTip('Define the % volume fraction of air bubble inclusions in the matrix')
        form.addRow(label, self.bubble_vf_sb)
        #
        # Bubble radius in microns
        #
        self.bubble_radius_sb = QDoubleSpinBox(self)
        self.bubble_radius_sb.setRange(0.001, 1000.0)
        self.bubble_radius_sb.setSingleStep(1.0)
        self.bubble_radius_sb.setDecimals(3)
        self.bubble_radius_sb.setToolTip('Define the air bubble radius')
        self.bubble_radius_sb.setValue(self.settings['Bubble radius'])
        self.bubble_radius_sb.valueChanged.connect(self.on_bubble_radius_sb_changed)
        label = QLabel('Air void radius (μm)', self)
        label.setToolTip('Define the air void radius')
        form.addRow(label, self.bubble_radius_sb)
        #
        # Mass fraction of dielectric medium
        #
        self.mf_sb = QDoubleSpinBox(self)
        self.mf_sb.setRange(0.000001, 100.0)
        self.mf_sb.setSingleStep(0.1)
        self.mf_sb.setDecimals(6)
        self.mf_sb.setToolTip('The percentage mass fraction of the dielectric medium. \nNote that volume and mass fraction are linked')
        self.mf_sb.setValue(100.0*self.settings['Mass fraction'])
        self.mf_sb.valueChanged.connect(self.on_mf_sb_changed)
        label = QLabel('% Mass fraction of dielectric', self)
        label.setToolTip('The percentage mass fraction of the dielectric medium. \nNote that volume and mass fraction are linked')
        form.addRow(label, self.mf_sb)
        #
        # Volume fraction of dielectric medium
        #
        self.vf_sb = QDoubleSpinBox(self)
        self.vf_sb.setRange(0.000001, 100.0*(1.0-self.settings['Bubble volume fraction']))
        self.vf_sb.setSingleStep(0.1)
        self.vf_sb.setDecimals(6)
        self.vf_sb.setToolTip('The percentage volume fraction of the dielectric medium. \nNote that volume and mass fraction are linked')
        self.vf_sb.valueChanged.connect(self.on_vf_sb_changed)
        self.vf_sb.setValue(100.0*self.settings['Volume fraction'])
        label = QLabel('% Volume fraction of dielectric', self)
        label.setToolTip('The percentage volume fraction of the dielectric medium. \nNote that volume and mass fraction are linked')
        form.addRow(label, self.vf_sb)
        #
        # Calculation method
        #
        self.methods_cb = QComboBox(self)
        self.methods_cb.setToolTip('Choose the calculation method for the effective medium theory')
        self.methods_cb.addItems(self.methods)
        index = self.methods_cb.findText(self.settings['Effective medium method'], Qt.MatchFixedString)
        if index >=0:
            self.methods_cb.setCurrentIndex(index)
        else:
            print('Method index was not 0',self.settings['Effective medium method'])
        self.methods_cb.activated.connect(self.on_methods_cb_activated)
        label = QLabel('Method',self)
        label.setToolTip('Choose the calculation method for the effective medium theory')
        form.addRow(label, self.methods_cb)
        #
        # Particle size option
        #
        self.size_sb = QDoubleSpinBox(self)
        self.size_sb.setRange(0.000001, 1000.0)
        self.size_sb.setSingleStep(0.1)
        self.size_sb.setDecimals(6)
        self.size_sb.setToolTip('Define the particle radius of the sphere in μm.')
        self.size_sb.setValue(self.settings['Particle size(mu)'])
        self.size_sb.valueChanged.connect(self.on_size_sb_changed)
        label = QLabel('Particle radius (μm)',self)
        label.setToolTip('Define the particle radius of the sphere in μm.')
        form.addRow(label, self.size_sb)
        #
        # Particle sigma option
        #
        self.sigma_sb = QDoubleSpinBox(self)
        self.sigma_sb.setRange(0.0, 1000.0)
        self.sigma_sb.setSingleStep(0.1)
        self.sigma_sb.setDecimals(6)
        self.sigma_sb.setToolTip('Define the particle size distribution as a lognormal distribution with the given sigma. \nOnly applicable for the Mie method')
        self.sigma_sb.setValue(self.settings['Particle size distribution sigma(mu)'])
        self.sigma_sb.valueChanged.connect(self.on_sigma_sb_changed)
        label = QLabel('Particle sigma (μm)',self)
        label.setToolTip('Define the particle size distribition as a lognormal with the given sigma. \nOnly applicable for the Mie method')
        form.addRow(label, self.sigma_sb)
        #
        # Crystallite shape
        #
        self.shape_cb = QComboBox(self)
        self.shape_cb.setToolTip('Choose a particle shape. \nFor the Mie methods only sphere is allowed.  \nFor shapes other than sphere there is a unique direction. \nFor ellipsoidal and needle like this is a direction [abc].  \nFor a plate the perpendicular to a crystal face (hkl) is used to define the unique direction')
        self.shape_cb.addItems(self.shapes)
        index = self.shape_cb.findText(self.settings['Particle shape'], Qt.MatchFixedString)
        if index >=0:
            self.shape_cb.setCurrentIndex(index)
        else:
            print('Method index was not 0',self.settings['Particle shape'])
        self.shape_cb.activated.connect(self.on_shape_cb_activated)
        label = QLabel('Particle shape',self)
        label.setToolTip('Choose a particle shape. \nFor the Mie methods only sphere is allowed.  \nFor shapes other than sphere there is a unique direction. \nFor ellipsoidal and needle like this is a direction [abc].  \nFor a plate the perpendicular to a crystal face (hkl) is used to define the unique direction')
        form.addRow(label, self.shape_cb)
        #
        # Particle shape information
        # unique direction (hkl) or [abc]
        self.h_sb = QSpinBox(self)
        self.h_sb.setToolTip('Define the h dimension of the unique direction')
        self.h_sb.setRange(-20,20)
        self.h_sb.setValue(self.settings['Unique direction - h'])
        self.h_sb.valueChanged.connect(self.on_h_sb_changed)
        self.k_sb = QSpinBox(self)
        self.k_sb.setToolTip('Define the k dimension of the unique direction')
        self.k_sb.setRange(-20,20)
        self.k_sb.setValue(self.settings['Unique direction - k'])
        self.k_sb.valueChanged.connect(self.on_k_sb_changed)
        self.l_sb = QSpinBox(self)
        self.l_sb.setToolTip('Define the l dimension of the unique direction')
        self.l_sb.setRange(-20,20)
        self.l_sb.setValue(self.settings['Unique direction - l'])
        self.l_sb.valueChanged.connect(self.on_l_sb_changed)
        hbox = QHBoxLayout()
        hbox.addWidget(self.h_sb)
        hbox.addWidget(self.k_sb)
        hbox.addWidget(self.l_sb)
        self.hkl_label = QLabel('Unique direction [abc]',self)
        self.hkl_label.setToolTip('Define the unique direction by [abc] or (hkl). \n[abc] is used by needles and ellipsoids.  It defines the unique direction in crystallographic units. \n(hkl) is used by plates it defines a surface and the unique direction is perpendicular to it.')

        form.addRow(self.hkl_label, hbox)
        #
        # a over b ratio for ellipse
        #
        self.aoverb_sb = QDoubleSpinBox(self)
        self.aoverb_sb.setRange(0.0, 1000.0)
        self.aoverb_sb.setSingleStep(0.1)
        self.aoverb_sb.setDecimals(6)
        self.aoverb_sb.setToolTip('Define the ellipsoid a/b ratio or eccentricity.  \nOnly applicable for the ellipsoid shapes \na/b < 1: oblate ellipsoid \na/b > 1: prolate ellipsoid')
        self.aoverb_sb.setValue(self.settings['Ellipsoid a/b'])
        self.aoverb_sb.valueChanged.connect(self.on_aoverb_sb_changed)
        label = QLabel('Ellipsoid a/b eccentricty',self)
        label.setToolTip('Define the ellipsoid a/b ratio or eccentricity.  \nOnly applicable for the ellipsoid shapes \na/b < 1: oblate ellipsoid \na/b > 1: prolate ellipsoid')
        form.addRow(label, self.aoverb_sb)
        #
        # Add ATR options
        # Refractive Index
        self.atr_index_sb = QDoubleSpinBox(self)
        self.atr_index_sb.setRange(0.001, 100.0)
        self.atr_index_sb.setSingleStep(0.01)
        self.atr_index_sb.setDecimals(3)
        self.atr_index_sb.setToolTip('Define the ATR material refractive index')
        self.atr_index_sb.setValue(self.settings['ATR material refractive index'])
        self.atr_index_sb.valueChanged.connect(self.on_atr_index_sb_changed)
        label = QLabel('ATR material refractive index', self)
        label.setToolTip('Define the ATR material refractive index')
        form.addRow(label, self.atr_index_sb)
        # Incident angle in degreees
        self.atr_incident_ang_sb = QDoubleSpinBox(self)
        self.atr_incident_ang_sb.setRange(0.0, 180.0)
        self.atr_incident_ang_sb.setSingleStep(0.1)
        self.atr_incident_ang_sb.setDecimals(1)
        self.atr_incident_ang_sb.setToolTip('Define the ATR incident angle')
        self.atr_incident_ang_sb.setValue(self.settings['ATR theta'])
        self.atr_incident_ang_sb.valueChanged.connect(self.on_atr_incident_ang_sb_changed)
        label = QLabel('ATR incident angle', self)
        label.setToolTip('Define the ATR incident angle')
        form.addRow(label, self.atr_incident_ang_sb)
        # S polarisation fraction
        self.atr_spolfrac_sb = QDoubleSpinBox(self)
        self.atr_spolfrac_sb.setRange(0.0, 1.0)
        self.atr_spolfrac_sb.setSingleStep(0.01)
        self.atr_spolfrac_sb.setDecimals(3)
        self.atr_spolfrac_sb.setToolTip('Define the ATR S polarisation fraction, the rest is P polarisation')
        self.atr_spolfrac_sb.setValue(self.settings['ATR S polarisation fraction'])
        self.atr_spolfrac_sb.valueChanged.connect(self.on_atr_spolfrac_sb_changed)
        label = QLabel('ATR S polarisation fraction', self)
        label.setToolTip('Define the S polarisation fraction, the rest is P polarisation')
        form.addRow(label, self.atr_spolfrac_sb)
        #
        # Add a legend option
        #
        self.legend_le = QLineEdit(self)
        self.legend_le.setToolTip('The legend will be used to describe the results in the plot')
        self.legend_le.setText('Scenario legend')
        self.legend_le.textChanged.connect(self.on_legend_le_changed)
        label = QLabel('Scenario legend',self)
        label.setToolTip('The legend will be used to describe the results in the plot')
        form.addRow(label, self.legend_le)

        #
        # Final buttons
        #
        hbox = QHBoxLayout()
        self.pushButton1 = QPushButton('Add another scenario')
        self.pushButton1.setToolTip('Use another scenario to calculate the effect of changing the material on the absorption and permittivity')
        self.pushButton1.clicked.connect(self.pushButton1Clicked)
        hbox.addWidget(self.pushButton1)
        self.pushButton3 = QPushButton('Delete this scenario')
        self.pushButton3.setToolTip('Delete the current scenario')
        self.pushButton3.clicked.connect(self.pushButton3Clicked)
        hbox.addWidget(self.pushButton3)
        form.addRow(hbox)
        vbox.addLayout(form)
        # finalise the layout
        self.setLayout(vbox)
        # sort out greying of boxes
        self.change_greyed_out()

    def pushButton1Clicked(self):
        # Add another scenario
        debugger.print('Button 1 pressed')
        self.notebook.addScenario(copyFromIndex=self.scenarioIndex)

    def pushButton3Clicked(self):
        # Delete a scenario
        debugger.print('Button 3 pressed')
        self.notebook.deleteScenario(self.scenarioIndex)

    def crystal_density(self):
        if not self.reader:
            return 1.0
        volume = self.reader.volume
        mass = 0.0
        for m in self.reader.masses:
            mass += m
        density = mass / (avogadro_si * volume * 1.0e-24)
        return density


    def on_h_sb_changed(self,value):
        debugger.print('on_h_sb_changed', value)
        self.dirty = True
        self.notebook.plottingCalculationRequired = True
        self.notebook.fittingCalculationRequired = True
        self.settings['Unique direction - h'] = value

    def on_k_sb_changed(self,value):
        debugger.print('on_k_sb_changed', value)
        self.dirty = True
        self.notebook.plottingCalculationRequired = True
        self.notebook.fittingCalculationRequired = True
        self.settings['Unique direction - k'] = value

    def on_l_sb_changed(self,value):
        debugger.print('on_l_sb_changed', value)
        self.dirty = True
        self.notebook.plottingCalculationRequired = True
        self.notebook.fittingCalculationRequired = True
        self.settings['Unique direction - l'] = value

    def on_shape_cb_activated(self,index):
        debugger.print('on shape cb activated', index)
        self.dirty = True
        self.notebook.plottingCalculationRequired = True
        self.notebook.fittingCalculationRequired = True
        self.settings['Particle shape'] = self.shapes[index]
        if self.settings['Particle shape'] == 'Sphere':
            self.settings['Unique direction - h'] = 0
            self.settings['Unique direction - k'] = 0
            self.settings['Unique direction - l'] = 0
        self.change_greyed_out()

    def on_methods_cb_activated(self,index):
        debugger.print('on methods cb activated', index)
        self.dirty = True
        self.notebook.plottingCalculationRequired = True
        self.notebook.fittingCalculationRequired = True
        self.settings['Effective medium method'] = self.methods[index]
        if self.settings['Effective medium method'] == 'Mie':
            self.settings['Particle shape'] = 'Sphere'
        elif self.settings['Effective medium method'] == 'Anisotropic-Mie':
            self.settings['Particle shape'] = 'Sphere'
        elif self.settings['Effective medium method'] == 'Maxwell-Garnett':
            self.settings['Particle size distribution sigma(mu)'] = 0.0
        elif self.settings['Effective medium method'] == 'Bruggeman':
            self.settings['Particle size distribution sigma(mu)'] = 0.0
        elif self.settings['Effective medium method'] == 'Averaged Permittivity':
            self.settings['Particle size(mu)'] = 0.0001
            self.settings['Particle size distribution sigma(mu)'] = 0.0
        self.change_greyed_out()

    def on_mf_sb_changed(self,value):
        debugger.print('on mass fraction line edit changed', value)
        self.dirty = True
        self.notebook.plottingCalculationRequired = True
        self.notebook.fittingCalculationRequired = True
        self.settings['Mass or volume fraction'] = 'mass'
        self.settings['Mass fraction'] =  value/100.0
        self.update_vf_sb()

    def update_vf_sb(self):
        mf1 = self.settings['Mass fraction']
        mf2 = 1.0 - mf1
        rho1 = self.crystal_density()
        rho2 = self.settings['Matrix density']
        vf1 = ( 1.0 - self.settings['Bubble volume fraction'] ) * (mf1/mf2)*(rho2/rho1) / ( 1 + (mf1/mf2)*(rho2/rho1))
#        vf1 = 1.0 / ( 1.0 + mf2/mf1 * (rho1/rho2) )
        self.settings['Volume fraction'] = vf1
        self.vf_sb.blockSignals(True)
        self.vf_sb.setValue(100.0*vf1)
        self.vf_sb.blockSignals(False)
        self.bubble_vf_sb.setRange(0.0, 100.0*(1.0-self.settings['Volume fraction']))
        self.vf_sb.setRange(0.0, 100.0*(1.0-self.settings['Bubble volume fraction']))
        debugger.print('Update_vf_sb')
        debugger.print('rho 1', rho1)
        debugger.print('rho 2', rho2)
        debugger.print('vf 1 ', vf1)

    def on_aoverb_sb_changed(self,value):
        debugger.print('on_aoverb_le_changed',value)
        self.dirty = True
        self.notebook.plottingCalculationRequired = True
        self.notebook.fittingCalculationRequired = True
        self.settings['Ellipsoid a/b'] = value

    def on_legend_le_changed(self,text):
        debugger.print('on legend change', text)
        self.dirty = True
        self.settings['Legend'] = text

    def on_sigma_sb_changed(self,value):
        debugger.print('on sigma line edit changed', value)
        self.dirty = True
        self.notebook.plottingCalculationRequired = True
        self.notebook.fittingCalculationRequired = True
        self.settings['Particle size distribution sigma(mu)'] = value

    def on_size_sb_changed(self,value):
        debugger.print('on size line edit changed', value)
        self.dirty = True
        self.notebook.plottingCalculationRequired = True
        self.notebook.fittingCalculationRequired = True
        self.settings['Particle size(mu)'] = value

    def on_vf_sb_changed(self,value):
        debugger.print('on volume fraction line edit changed', value)
        self.dirty = True
        self.notebook.plottingCalculationRequired = True
        self.notebook.fittingCalculationRequired = True
        self.settings['Mass or volume fraction'] = 'volume'
        self.settings['Volume fraction'] = value/100.0
        self.update_mf_sb()

    def update_mf_sb(self):
        vf1 = self.settings['Volume fraction']
        vf2 = 1.0 - vf1 - self.settings['Bubble volume fraction']
        rho1 = self.crystal_density()
        rho2 = self.settings['Matrix density']
        # mf1 = 1.0 / ( 1.0 + (vf2/vf1) * (rho2/rho1) )
        mf1 = rho1*vf1 / ( rho1*vf1 + rho2*vf2 )
        self.settings['Mass fraction'] = mf1
        self.mf_sb.blockSignals(True)
        self.mf_sb.setValue(100.0*mf1)
        self.mf_sb.blockSignals(False)
        debugger.print('Update_mf_sb')
        debugger.print('rho 1', rho1)
        debugger.print('rho 2', rho2)
        debugger.print('mf 1 ', mf1)

    def on_matrix_cb_activated(self,index):
        debugger.print('on matrix combobox activated', index)
        debugger.print('on matrix combobox activated', self.matrix_cb.currentText())
        self.dirty = True
        self.notebook.plottingCalculationRequired = True
        self.notebook.fittingCalculationRequired = True
        matrix = self.matrix_cb.currentText()
        self.matrix_cb.blockSignals(True)
        self.density_sb.blockSignals(True)
        self.permittivity_sb.blockSignals(True)
        self.settings['Matrix'] = matrix
        self.settings['Matrix density'] = support_matrix_db[matrix][0]
        self.settings['Matrix permittivity'] = support_matrix_db[matrix][1]
        self.density_sb.setValue(self.settings['Matrix density'])
        self.permittivity_sb.setValue(self.settings['Matrix permittivity'])
        # volume fraction takes precedence
        if self.settings['Mass or volume fraction'] == 'volume':
            self.update_mf_sb()
            self.update_vf_sb()
        else:
            self.update_vf_sb()
            self.update_mf_sb()
        self.matrix_cb.blockSignals(False)
        self.density_sb.blockSignals(False)
        self.permittivity_sb.blockSignals(False)

    def on_density_sb_changed(self,value):
        self.settings['Matrix density'] = value
        # volume fraction taked precedence
        if self.settings['Mass or volume fraction'] == 'volume':
            self.update_mf_sb()
            self.update_vf_sb()
        else:
            self.update_vf_sb()
            self.update_mf_sb()
        debugger.print('on density line edit changed', value)
        self.dirty = True
        self.notebook.plottingCalculationRequired = True
        self.notebook.fittingCalculationRequired = True

    def on_bubble_vf_sb_changed(self,value):
        self.settings['Bubble volume fraction'] = value/100.0
        if self.settings['Mass or volume fraction'] == 'volume':
            self.update_mf_sb()
        else:
            self.update_vf_sb()
        debugger.print('on bubble volume fraction changed', value)
        self.dirty = True
        self.notebook.plottingCalculationRequired = True
        self.notebook.fittingCalculationRequired = True

    def on_bubble_radius_sb_changed(self,value):
        self.settings['Bubble radius'] = value
        debugger.print('on permittivity line edit changed', value)
        self.dirty = True
        self.notebook.plottingCalculationRequired = True
        self.notebook.fittingCalculationRequired = True

    def on_permittivity_sb_changed(self,value):
        self.settings['Matrix permittivity'] = value
        debugger.print('on permittivity line edit changed', value)
        self.dirty = True
        self.notebook.plottingCalculationRequired = True
        self.notebook.fittingCalculationRequired = True

    def on_atr_index_sb_changed(self,value):
        self.settings['ATR material refractive index'] = value
        debugger.print('on atr index line edit changed', value)
        self.dirty = True
        self.notebook.plottingCalculationRequired = True
        self.notebook.fittingCalculationRequired = True

    def on_atr_incident_ang_sb_changed(self,value):
        self.settings['ATR theta'] = value
        debugger.print('on atr incident angle line edit changed', value)
        self.dirty = True
        self.notebook.plottingCalculationRequired = True
        self.notebook.fittingCalculationRequired = True

    def on_atr_spolfrac_sb_changed(self,value):
        self.settings['ATR S polarisation fraction'] = value
        debugger.print('on atr spolfraction line edit changed', value)
        self.dirty = True
        self.notebook.plottingCalculationRequired = True
        self.notebook.fittingCalculationRequired = True

    def set_reader(self,reader):
        self.dirty = True
        self.notebook.plottingCalculationRequired = True
        self.notebook.fittingCalculationRequired = True
        self.reader = reader

    def change_greyed_out(self):
        # Have a look through the settings and see if we need to grey anything out
        method = self.settings['Effective medium method']
        if method == 'Mie' or method == 'Anisotropic-Mie':
            self.size_sb.setEnabled(True)
            self.sigma_sb.setEnabled(True)
            for i,shape in enumerate(self.shapes):
                self.shape_cb.model().item(i).setEnabled(False)
            self.settings['Particle shape'] = 'Sphere'
            self.shape_cb.setEnabled(True)
            index = self.shape_cb.findText(self.settings['Particle shape'], Qt.MatchFixedString)
            if index >=0:
                self.shape_cb.model().item(index).setEnabled(True)
                self.shape_cb.setCurrentIndex(index)
            else:
                print('Method index was not 0',self.settings['Particle shape'])
        elif method == 'Averaged Permittivity':
            self.size_sb.setEnabled(False)
            self.sigma_sb.setEnabled(False)
            self.settings['Particle shape'] = 'Sphere'
            index = self.shape_cb.findText(self.settings['Particle shape'], Qt.MatchFixedString)
            if index >=0:
                self.shape_cb.model().item(index).setEnabled(True)
                self.shape_cb.setCurrentIndex(index)
            self.shape_cb.setEnabled(False)
            for i,shape in enumerate(self.shapes):
                self.shape_cb.model().item(i).setEnabled(False)
        elif method == 'Maxwell-Garnett' or method == 'Bruggeman':
            self.size_sb.setEnabled(True)
            self.sigma_sb.setEnabled(False)
            self.shape_cb.setEnabled(True)
            for i,shape in enumerate(self.shapes):
                self.shape_cb.model().item(i).setEnabled(True)
        else:
            self.size_sb.setEnabled(False)
            self.sigma_sb.setEnabled(False)
            self.shape_cb.setEnabled(True)
            for i,shape in enumerate(self.shapes):
                self.shape_cb.model().item(i).setEnabled(True)
        # deal with shapes
        if self.settings['Particle shape'] == 'Ellipsoid':
            self.h_sb.setEnabled(True)
            self.k_sb.setEnabled(True)
            self.l_sb.setEnabled(True)
            self.hkl_label.setText('Unique direction [abc]')
            self.aoverb_sb.setEnabled(True)
        elif self.settings['Particle shape'] == 'Plate':
            self.h_sb.setEnabled(True)
            self.k_sb.setEnabled(True)
            self.l_sb.setEnabled(True)
            self.hkl_label.setText('Unique direction (hkl)')
            self.aoverb_sb.setEnabled(False)
        elif self.settings['Particle shape'] == 'Needle':
            self.h_sb.setEnabled(True)
            self.k_sb.setEnabled(True)
            self.l_sb.setEnabled(True)
            self.hkl_label.setText('Unique direction [abc]')
            self.aoverb_sb.setEnabled(False)
        elif self.settings['Particle shape'] == 'Sphere':
            self.h_sb.setEnabled(False)
            self.k_sb.setEnabled(False)
            self.l_sb.setEnabled(False)
            self.aoverb_sb.setEnabled(False)
        else:
            print('ScenarioTab: Shape not recognised', self.settings['Particle shape'])

    def setScenarioIndex(self,index):
        self.scenarioIndex = index
        text = self.legend_le.text()
        if text == 'Scenario legend':
            self.legend_le.setText('Scenario '+str(index + 1))
        return

    def print_settings(self):
        print('#')
        print('# Scenario tab')
        print('#')
        print('tab = self.notebook.scenarios')
        for key in self.settings:
            print(key, self.settings[key])

    def refresh(self,force=False):
        if not self.dirty and not force:
            debugger.print('refresh aborted', self.dirty,force)
            return
        debugger.print('refresh', force)
        # Tell the main notebook that we need to recalculate any plot
        self.notebook.plottingCalculationRequired = True
        self.notebook.fittingCalculationRequired = True
        # First see if we can get the reader from the mainTab
        self.reader = self.notebook.mainTab.reader
        #
        # Block signals during refresh
        #
        for w in self.findChildren(QWidget):
            w.blockSignals(True)
        # use the settings values to initialise the widgets
        index = self.matrix_cb.findText(self.settings['Matrix'], Qt.MatchFixedString)
        self.matrix_cb.setCurrentIndex(index)
        self.density_sb.setValue(self.settings['Matrix density'])
        self.permittivity_sb.setValue(self.settings['Matrix permittivity'])
        self.bubble_vf_sb.setValue(100*self.settings['Bubble volume fraction'])
        self.bubble_radius_sb.setValue(self.settings['Bubble radius'])
        if self.settings['Mass or volume fraction'] == 'volume':
            # volume fraction takes precedence
            self.update_mf_sb()
            self.update_vf_sb()
        else:
            # mass fraction takes precedence
            self.update_vf_sb()
            self.update_mf_sb()
        #
        index = self.methods_cb.findText(self.settings['Effective medium method'], Qt.MatchFixedString)
        self.methods_cb.setCurrentIndex(index)
        self.size_sb.setValue(self.settings['Particle size(mu)'])
        self.sigma_sb.setValue(self.settings['Particle size distribution sigma(mu)'])
        index = self.shape_cb.findText(self.settings['Particle shape'], Qt.MatchFixedString)
        self.shape_cb.setCurrentIndex(index)
        self.h_sb.setValue(self.settings['Unique direction - h'])
        self.k_sb.setValue(self.settings['Unique direction - k'])
        self.l_sb.setValue(self.settings['Unique direction - l'])
        self.aoverb_sb.setValue(self.settings['Ellipsoid a/b'])
        self.legend_le.setText(self.settings['Legend'])
        self.change_greyed_out()
        #
        # Unblock signals after refresh
        #
        for w in self.findChildren(QWidget):
            w.blockSignals(False)
        self.dirty = False
        return



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="PDielec",
    version="6.3.0",
    author="John Kendrick and Andrew Burnett",
    author_email="john@kendrick.co.uk",
    description="PDielec package for the calculation of THz and infrared spectra",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/JohnKendrick/PDielec",
    packages=setuptools.find_packages(include=['PDielec','PDielec.*']),
    license='MIT',
    keywords=['Infrared',' Infrared Spectroscopy','THz Spectroscopy','Terahertz Spectroscopy','Bruggeman','Maxwell-Garnett','Effective Medium'],
    install_requires=[
        'numpy',
        'scipy',
        'PyYAML',
        'matplotlib',
        'termcolor',
        'xlsxwriter',
        'openpyxl',
        'imageio',
        'PyMieScatt',
    ],
    include_package_data=True,   # Assume all information is in MANIFEST.in
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=2.7',
    entry_points= {
        'console_scripts': [
            'preader=preader:main',
        ],
        'gui_scripts': [
            'pdgui=pdgui:main',
        ],
    },
)




from traits.api import HasTraits, Button, Range, Float, Int, Instance, on_trait_change, Property
from traitsui.api import View, Item, Group, HGroup, RangeEditor, Label, spring
from mayavi.core.api import PipelineBase
from mayavi.core.ui.api import MayaviScene, SceneEditor, MlabSceneModel
import numpy as np
from mayavi import mlab
from PDielec.Constants import covalent_radii, elemental_colours

class ViewerClass(HasTraits):
    number_of_modesm1 = 0
    maxdisplacement = 0.3
    gui_mode_selection    = Int(label="Mode")
    gui_maxdisplacement   = Float(maxdisplacement,label="Displacement")
    gui_savepicture_button = Button('Snapshot')
    gui_animate_button    = Button('Animate')
    gui_exit_button       = Button('Exit')
    # gui_exit_button.label = '?????'
    #from objbrowser import browse
    #browse(gui_exit_button)
    selected_mode_frequency = Property(Float, depends_on = ['gui_mode_selection'], label="Fequency")
#
    scene = Instance(MlabSceneModel, ())
    plot = Instance(PipelineBase)
    view = View(Item('scene', editor=SceneEditor(scene_class=MayaviScene),
                     height=250, width=300, show_label=False,resizable=True),
                Group(
                    Item('gui_mode_selection',
                         editor =  RangeEditor(low=0, high=number_of_modesm1, high_name='number_of_modesm1',format='%d',mode='auto')
                        ),
#jk                    Label('Select mode number for display of displacements'),
                    Item('selected_mode_frequency',style='readonly'),
                    'gui_maxdisplacement',
                    label='Mode selector',
                    show_labels = True,

                ),
                HGroup(
#                    'gui_savepicture_button',
                    Item('_'),
                    spring,
                    'gui_animate_button',
                    spring,
                    'gui_exit_button',
                    spring,
                    show_labels = False,
                ),
		resizable=True,
           )


    @on_trait_change('scene.activated')
    def update_plot(self):
        #jk print('update_plot',self.gui_mode_selection)
        if self.plot is None:
            #jk print('plot is None')
            self.draw()
        else:
            #jk print('plot None')
            pass

    @on_trait_change('gui_animate_button')
    def update_plot_on_animate_button(self):
        #print('Gui animate button has been pressed ')
        if self.animation:
           self.animation = False
           self.draw()
        else:
           self.animation = True
           self.animate()

    @on_trait_change('gui_savepicture_button')
    def update_plot_on_savepicture(self):
        print('Gui savepicture button has been pressed ')
        filename = 'snapshot_'+str(self.snapshot_number)+'.pdf'
        print(filename)
        mlab.savefig(filename, figure=mlab.gcf(),magnification=3)


    @on_trait_change('gui_maxdisplacement')
    def update_plot_on_maxdisplacement(self):
        #jk print('Gui maxdisplacement button has been pressed ')
        self.maxdisplacement = self.gui_maxdisplacement
        self.calculatePhasePositions()

    @on_trait_change('gui_exit_button')
    def update_plot_on_exit_button(self):
        #print('Gui exit button has been pressed ')
        exit()

    @on_trait_change('gui_mode_selection')
    def update_plot_on_mode_change(self):
        #print('mode has changed ',self.gui_mode_selection)
        self.mode = self.gui_mode_selection
        self.calculatePhasePositions()
        self.refreshDisplacements()

    def __init__(self, unit_cell, normal_modes, frequencies):
        super(ViewerClass, self).__init__()
        self.quiverBonds = False
        self.snapshot_number = 0
        self.unit_cell = unit_cell
        self.normal_modes = normal_modes
        self.frequencies = frequencies
        self.bonds = unit_cell.bonds
        self.nbonds = len(self.bonds)
        self.scale_vibrations = 20.0
        self.animation = False
        self.X = self.unit_cell.xyz_coordinates[:,0]
        self.Y = self.unit_cell.xyz_coordinates[:,1]
        self.Z = self.unit_cell.xyz_coordinates[:,2]
        self.natoms = len(self.X)
        self.number_of_modes = len(normal_modes)
        self.number_of_modesm1 = len(normal_modes)-1
        # reorder the displacement info in the normal modes into U,V and W lists
        self.U, self.V, self.W = [], [], []
        for mode,displacements in enumerate(self.normal_modes):
            U = np.zeros_like( self.X )
            V = np.zeros_like( self.X )
            W = np.zeros_like( self.X )
            index = -1
            for i in range(0,len(displacements),3):
                index += 1
                U[index] = displacements[i + 0]
                V[index] = displacements[i + 1]
                W[index] = displacements[i + 2]
            self.U.append(U)
            self.V.append(V)
            self.W.append(W)
        # get the cell edges for the bounding box
        corners,self.cell_edges = self.unit_cell.getBoundingBox()
        self.element_names = self.unit_cell.element_names
        self.colours = []
        self.radii = []
        # set the colours (the colour is a triplet of floats from 0.0 to 1.0)
        for el in self.element_names:
           colour_list = [ colour/255.0 for colour in elemental_colours[el] ]
           colour_tuple = tuple(colour_list)
           self.colours.append( colour_tuple )
           self.radii.append(1.3*covalent_radii[el])
        # We need these glyphs to update the atom and bond positions for animations
        self.glyph_atoms = []
        self.glyph_bonds = []
        self.mode = 0
        self.number_of_phase_steps = 21
        self.phase_direction = 1
        self.phase_index = 0
        self.calculatePhasePositions()
        return

    def calculatePhasePositions(self):
        # we need the number of phase steps to be odd
        if self.number_of_phase_steps%2 == 0:
            self.numbers_of_phase_steps += 1
        U = self.U[self.mode]
        V = self.V[self.mode]
        W = self.W[self.mode]
        maxR = np.amax(U)
        maxR = max(maxR, np.amax(V) )
        maxR = max(maxR, np.amax(W))
        self.scale_vibrations = self.maxdisplacement / maxR
        #print('MaxR is',maxR)
        self.newX    = np.zeros( (self.number_of_phase_steps,self.natoms) )
        self.newY    = np.zeros( (self.number_of_phase_steps,self.natoms) )
        self.newZ    = np.zeros( (self.number_of_phase_steps,self.natoms) )
        self.startX  = np.zeros( (self.number_of_phase_steps,self.nbonds) )
        self.startY  = np.zeros( (self.number_of_phase_steps,self.nbonds) )
        self.startZ  = np.zeros( (self.number_of_phase_steps,self.nbonds) )
        self.dirX    = np.zeros( (self.number_of_phase_steps,self.nbonds) )
        self.dirY    = np.zeros( (self.number_of_phase_steps,self.nbonds) )
        self.dirZ    = np.zeros( (self.number_of_phase_steps,self.nbonds) )
        self.bondx   = np.zeros( (self.number_of_phase_steps,self.natoms,2) )
        self.bondy   = np.zeros( (self.number_of_phase_steps,self.natoms,2) )
        self.bondz   = np.zeros( (self.number_of_phase_steps,self.natoms,2) )
        # work out the scaling factor to get the require maximumdisplacement
        sign = +1.0
        phases = []
        small = 1.0E-8
        n2 = int(self.number_of_phase_steps/2)
        delta = 1.0 / n2
        phase = -delta
        for phase_index in range(self.number_of_phase_steps):
            phase += delta*sign
            self.newX[phase_index] = self.X+U*phase*self.scale_vibrations
            self.newY[phase_index] = self.Y+V*phase*self.scale_vibrations
            self.newZ[phase_index] = self.Z+W*phase*self.scale_vibrations
            if phase >= 1.0-small:
                sign = -1.0
                phase += delta*sign
            elif phase <= -1.0+small:
                sign = +1.0
                phase += delta*sign
            phases.append(phase)
        # end for phase_index
        for phase_index,phase in enumerate(phases):
            for ibond,bond in enumerate(self.bonds):
                i,j = bond
                self.bondx[phase_index,ibond,0] = self.newX[phase_index,i]
                self.bondy[phase_index,ibond,0] = self.newY[phase_index,i]
                self.bondz[phase_index,ibond,0] = self.newZ[phase_index,i]
                self.bondx[phase_index,ibond,1] = self.newX[phase_index,j]
                self.bondy[phase_index,ibond,1] = self.newY[phase_index,j]
                self.bondz[phase_index,ibond,1] = self.newZ[phase_index,j]
                self.startX[phase_index,ibond] = self.newX[phase_index,i]
                self.startY[phase_index,ibond] = self.newY[phase_index,i]
                self.startZ[phase_index,ibond] = self.newZ[phase_index,i]
                self.dirX[phase_index,ibond] = self.newX[phase_index,j] - self.newX[phase_index,i]
                self.dirY[phase_index,ibond] = self.newY[phase_index,j] - self.newY[phase_index,i]
                self.dirZ[phase_index,ibond] = self.newZ[phase_index,j] - self.newZ[phase_index,i]
            #end for ibond,bond
        # end of phase_index,phase


    def animate(self):
        mlab.clf()
        self.draw_noupdate()
        animator = anim(self.phase_index, self.phase_direction, self.number_of_phase_steps, self.quiverBonds, self.refreshAtoms, self.refreshQuiverBonds,self.refreshBonds)
#        from objbrowser import browse
#        browse(animator)
#        print('In animate called anim')

    def refreshBonds(self, phase_index):
        for ig,glyph in enumerate(self.glyph_bonds):
            glyph.mlab_source.set(x=self.bondx[phase_index,ig], y=self.bondy[phase_index,ig], z=self.bondz[phase_index,ig])
        return

    def refreshQuiverBonds(self, phase_index):
        self.glyph_quiver_bonds.mlab_source.set(x=self.startX[phase_index],
                                                y=self.startY[phase_index],
                                                z=self.startZ[phase_index],
                                                u=self.dirX[phase_index],
                                                v=self.dirY[phase_index],
                                                w=self.dirZ[phase_index])
        return

    def refreshAtoms(self, phase_index):
        for ig,glyph in enumerate(self.glyph_atoms):
            glyph.mlab_source.set(x=self.newX[phase_index,ig], y=self.newY[phase_index,ig], z=self.newZ[phase_index,ig])
        return

    def refreshDisplacements(self):
        mode = self.mode
        self.glyph_displacements.mlab_source.set(u=self.U[mode],v=self.V[mode],w=self.W[mode])
        return

    def draw(self):
        mlab.clf()
        self.draw_noupdate()
        self.drawDisplacements()

    def draw_noupdate(self):
        self.drawUnitCell()
        if self.quiverBonds:
            self.drawQuiverBonds()
        else:
            self.drawBonds()
        self.drawAtoms()
        # self.drawDisplacements()
        return

    def drawDisplacements(self):
        mode = self.mode
        self.glyph_displacements = mlab.quiver3d(self.X,self.Y,self.Z,self.U[mode],self.V[mode],self.W[mode],color=(0.15,0.15,0.15),mode='arrow' )
        return

    def drawUnitCell(self):
        bondx = np.zeros( 2 )
        bondy = np.zeros( 2 )
        bondz = np.zeros( 2 )
        ones  = np.array( (1.0,1.0) )
        for edge in self.cell_edges:
            p1,p2 =edge
            bondx[0] = p1[0]
            bondy[0] = p1[1]
            bondz[0] = p1[2]
            bondx[1] = p2[0]
            bondy[1] = p2[1]
            bondz[1] = p2[2]
            mlab.plot3d(bondx,bondy,bondz,ones, tube_radius=0.15, tube_sides=10, color=(1,0,0) )
        return

    def drawQuiverBonds(self):
        self.glyph_quiver_bonds = mlab.quiver3d(self.startX[0],self.startY[0],self.startZ[0],
                                                self.dirX[0],  self.dirY[0],  self.dirZ[0],
                                                color=(0,0,0), mode='cylinder')

    def drawBonds(self):
        bondx = np.zeros( 2 )
        bondy = np.zeros( 2 )
        bondz = np.zeros( 2 )
        ones  = np.array( (1.0,1.0) )
        self.glyph_bonds = []
        for bond in self.unit_cell.bonds:
            i,j = bond
            bondx[0] = self.X[i]
            bondy[0] = self.Y[i]
            bondz[0] = self.Z[i]
            bondx[1] = self.X[j]
            bondy[1] = self.Y[j]
            bondz[1] = self.Z[j]
            glyph = mlab.plot3d(bondx,bondy,bondz,ones,tube_radius=0.1, tube_sides=10, color=(0,0,0) )
            self.glyph_bonds.append(glyph)
        return

    def drawAtoms(self):
        self.glyph_atoms = []
        for x,y,z,colour,size in zip(self.X, self.Y, self.Z, self.colours, self.radii):
           glyph = mlab.points3d([x],[y],[z],[1],color=colour,resolution=20, scale_factor=size)
           self.glyph_atoms.append(glyph)
        return

    def _get_selected_mode_frequency(self):
        self.mode = self.gui_mode_selection
        self.calculatePhasePositions()
        return self.frequencies[self.gui_mode_selection]

@mlab.animate(delay=20, ui=True, support_movie=False)
def anim(phase_index, phase_direction, number_of_phase_steps, quiverBonds, refreshAtoms, refreshQuiverBonds, refreshBonds ):
    f = mlab.gcf()
    f.scene.anti_aliasing_frames = 0
    while True:
        phase_index = phase_index + phase_direction
        if phase_index >= int(number_of_phase_steps/2):
            phase_direction = -1
        elif phase_index <= -int(number_of_phase_steps/2):
            phase_direction = +1
        f.scene.disable_render = True
        if quiverBonds:
            refreshQuiverBonds(phase_index)
        else:
            refreshBonds(phase_index)
        refreshAtoms(phase_index)
        f.scene.disable_render = False
        yield
    return



#!/usr/bin/env python
"""Read the contents of a directory containing DFT output and create a cif file of the structure"""
from __future__ import print_function
import string
import re
import numpy as np
import os, sys
from PDielec.Constants import amu, PI, avogadro_si, wavenumber, angstrom, isotope_masses, average_masses
from PDielec.VaspOutputReader import VaspOutputReader
from PDielec.CastepOutputReader import CastepOutputReader
from PDielec.GulpOutputReader import GulpOutputReader
from PDielec.CrystalOutputReader import CrystalOutputReader
from PDielec.AbinitOutputReader import AbinitOutputReader
from PDielec.QEOutputReader import QEOutputReader
from PDielec.PhonopyOutputReader import PhonopyOutputReader
from multiprocessing import Pool, cpu_count
import PDielec.Calculator as Calculator

def set_affinity_on_worker():
    """When a new worker process is created, the affinity is set to all CPUs"""
    #JK print("I'm the process %d, setting affinity to all CPUs." % os.getpid())
    #JK for the time being this is simply commented out, but might be useful at some point
    #os.system("taskset -p 0xff %d > /dev/null" % os.getpid())

def read_a_file( calling_parameters):
    name, program, qmprogram, debug = calling_parameters
    fulldirname = name
    head,tail = os.path.split(fulldirname)
    root,ext = os.path.splitext(tail)
    if program == "castep":
        names = [ name ]
        reader = CastepOutputReader( names )
    elif program == "vasp":
        name1 = os.path.join(head,'OUTCAR')
        name2 = os.path.join(head,'KPOINTS')
        names = [ name1, name2 ]
        reader = VaspOutputReader( names )
    elif program == "gulp":
        names = [ name ]
        reader = GulpOutputReader( names )
    elif program == "crystal":
        names = [ name ]
        reader = CrystalOutputReader( names )
    elif program == "abinit":
        names = [ name ]
        reader = AbinitOutputReader( names )
    elif program == "qe":
        tail1 = root+'.dynG'
        tail2 = root+'.log'
        tail3 = root+'.out'
        name1 = os.path.join(head,tail2)
        name2 = os.path.join(head,tail3)
        name3 = os.path.join(head,tail)
        # We want the dynG entry last rounding causes problems otherwise
        name4 = os.path.join(head,tail1)
        names = []
        for n in [ name1, name2, name3, name4 ]:
            if os.path.isfile(n):
                if not n in names:
                    names.append(n)
        reader = QEOutputReader( names )
    elif program == "phonopy":
        # The order is important
        pname1 = os.path.join(head,'qpoints.yaml')
        pname2 = os.path.join(head,'phonopy.yaml')
        # Only works for VASP at the moment
        vname1 = os.path.join(head,'OUTCAR')
        vname2 = os.path.join(head,'KPOINTS')
        pnames = [ pname1, pname2 ]
        vnames = [ vname1, vname2 ]
        pnames.extend(vnames)
        names = pnames
        # Which QM program was used by PHONOPY?
        if qmprogram == "castep":
            print("Error in qmreader",qmprogram)
            exit()
            qmreader = CastepOutputReader(names)
        elif qmprogram == "vasp":
            qmreader = VaspOutputReader(vnames)
        elif qmprogram == "gulp":
            print("Error in qmreader",qmprogram)
            exit()
            qmreader = GulpOutputReader(names)
        elif qmprogram == "crystal":
            print("Error in qmreader",qmprogram)
            exit()
            qmreader = CrystalOutputReader(names)
        elif qmprogram == "abinit":
            print("Error in qmreader",qmprogram)
            exit()
            qmreader = AbinitOutputReader(names)
        elif qmprogram == "qe":
            print("Error in qmreader",qmprogram)
            exit()
            qmreader = QEOutputReader(names)
        # The QM reader is used to get info about the QM calculation
        reader = PhonopyOutputReader(pnames,qmreader)
    else:
        print('Program name not recognized',program,file=sys.stderr)
        exit()

    print('  Analysing ',names, file=sys.stderr)
    reader.debug = debug
    reader.read_output()
    return name,reader.unit_cells[-1]

def main():
    # Start processing the directories
    if len(sys.argv) <= 1 :
        print('p2cif -program program filenames .....', file=sys.stderr)
        print('  \"program\" must be one of \"abinit\", \"castep\", \"crystal\", \"gulp\"       ', file=sys.stderr)
        print('           \"phonopy\", \"qe\", \"vasp\"                                         ', file=sys.stderr)
        print('           If phonopy is used it must be followed by the QM package              ', file=sys.stderr)
        print('  -debug   to switch on more debug information                                   ', file=sys.stderr)
        exit()

    files = []
    tokens = sys.argv[1:]
    ntokens = len(tokens)-1
    itoken = -1
    program = ''
    qmprogram = ''
    debug = False
    while itoken < ntokens:
        itoken += 1
        token = tokens[itoken]
        if token == "-debug":
            debug = True
        elif token == "-program":
            itoken += 1
            program = tokens[itoken]
            if program == 'phonopy':
                itoken += 1
                qmprogram = tokens[itoken]
        else:
            files.append(token)

    if len(program) < 1:
        print('Please use -program to define the package used to generate the output files',file=sys.stderr)
        exit()

    if not program in ['abinit','castep','crystal','gulp','qe','vasp','phonopy']:
        print('Program is not recognised: ',program,file=sys.stderr)
        exit()

    if program == 'phonopy':
        if not qmprogram in ['abinit','castep','crystal','gulp','qe','vasp']:
            print('Phonopy QM program is not recognised: ',qmprogram,file=sys.stderr)
            exit()
        print('  QM program used by Phonopy is: ',qmprogram,file=sys.stderr)

    print('  Program is ',program,file=sys.stderr)

    for f in files:
        if not os.path.isfile(f):
            print('Error file requested for analysis does not exist',f,file=sys.stderr)
            exit()

    #
    # Create a pool of processors to handle reading the files
    #
    p = Pool(initializer=set_affinity_on_worker)
    # Create a tuple list of calling parameters
    calling_parameters = []
    files.sort()
    for name in files:
        calling_parameters.append( (name, program, qmprogram, debug) )
    # Calculate the results in parallel
    results_map_object = p.map_async(read_a_file,calling_parameters)
    results_map_object.wait()
    results = results_map_object.get()
    # Convert the results into a dictionary
    results_dictionary = {}
    for name,cell in results:
        cell.write_cif(filename=name,file_=sys.stdout)
    #
    exit()
# end of def main

if __name__ == "__main__":
    main(sys)



#!/usr/bin/python
#
# Copyright 2015 John Kendrick
#
# This file is part of PDielec
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the MIT License
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the MIT License
# along with this program, if not see https://opensource.org/licenses/MIT
#

"""Hold unit cell information and its associated calculated properties"""

from __future__ import print_function
import numpy as np
import math
from PDielec.Calculator  import calculate_distance
from PDielec.Plotter     import print_reals, print_ints, print_strings
from PDielec.Calculator  import cleanup_symbol
import sys


class UnitCell:
    """Hold unit cell information and its associated calculated properties"""
    def __init__(self, a=None, b=None, c=None, alpha=None, beta=None, gamma=None):
        self.fractional_coordinates = []
        self.xyz_coordinates = []
        self.element_names = []
        self.atom_labels = []
        self.bonds = []
        self.nions = 0
        self.molecules = []
        self.atomic_masses = []
        self.centres_of_mass = []
        self.total_mass = 0.0
        self.lattice = [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]
        if gamma is not None:
            self.lattice = self.convert_abc_to_unitcell(a, b, c, alpha, beta, gamma)
        else:
            self.lattice[0] = a
            self.lattice[1] = b
            self.lattice[2] = c
        self._calculate_reciprocal_lattice(self.lattice)

    def write_cif(self, filename='', file_=sys.stdout):
        print('data_'+filename,                                 file=file_)
        print('_space_group_IT_number 1',                       file=file_)
        print('_symmetry_space_group_name_H-M \'P1\'',          file=file_)
        print('_cell_length_a      {:12.6f}'.format(self.a),    file=file_)
        print('_cell_length_b      {:12.6f}'.format(self.b),    file=file_)
        print('_cell_length_c      {:12.6f}'.format(self.c),    file=file_)
        print('_cell_angle_alpha   {:12.6f}'.format(self.alpha),file=file_)
        print('_cell_angle_beta    {:12.6f}'.format(self.beta), file=file_)
        print('_cell_angle_gamma   {:12.6f}'.format(self.gamma),file=file_)
        print('loop_',                                          file=file_)
        print('_atom_site_label',                               file=file_)
        print('_atom_site_fract_x',                             file=file_)
        print('_atom_site_fract_y',                             file=file_)
        print('_atom_site_fract_z',                             file=file_)
        print('_atom_site_site_type_symbol',                    file=file_)
        for i,(frac,el) in enumerate(zip(self.fractional_coordinates,self.element_names)):
            symbol = el+str(i+1)
            print( '{} {:12.6f} {:12.6f} {:12.6f} {}'.format(symbol,frac[0], frac[1], frac[2], el), file=file_)
        return

    def getBoundingBox(self):
        # Return a box with 12 edges which represent the unit cell in cartesian space
        #  7---6
        # /|  /|
        # 3---2|
        # |4--|5
        # 0---1/
        # Corners in abc space
        corners = []
        corners.append( np.array( [0.0,0.0,0.0] ) )
        corners.append( np.array( [1.0,0.0,0.0] ) )
        corners.append( np.array( [1.0,1.0,0.0] ) )
        corners.append( np.array( [0.0,1.0,0.0] ) )
        corners.append( np.array( [0.0,0.0,1.0] ) )
        corners.append( np.array( [1.0,0.0,1.0] ) )
        corners.append( np.array( [1.0,1.0,1.0] ) )
        corners.append( np.array( [0.0,1.0,1.0] ) )
        # Now calculate the corners in cartesian space
        corners_xyz = [ self.convert_abc_to_xyz(corner) for corner in corners ]
        # Now calculate the edges each edge is a tuple with a beginning coordinate and an end coordinate
        edges = []
        edges.append( (corners_xyz[1] , corners_xyz[0]) )
        edges.append( (corners_xyz[2] , corners_xyz[1]) )
        edges.append( (corners_xyz[3] , corners_xyz[2]) )
        edges.append( (corners_xyz[3] , corners_xyz[0]) )
        edges.append( (corners_xyz[4] , corners_xyz[0]) )
        edges.append( (corners_xyz[5] , corners_xyz[1]) )
        edges.append( (corners_xyz[6] , corners_xyz[2]) )
        edges.append( (corners_xyz[7] , corners_xyz[3]) )
        edges.append( (corners_xyz[5] , corners_xyz[4]) )
        edges.append( (corners_xyz[6] , corners_xyz[5]) )
        edges.append( (corners_xyz[7] , corners_xyz[6]) )
        edges.append( (corners_xyz[7] , corners_xyz[4]) )
        return corners_xyz,edges


    def printInfo(self):
        print_reals('Unit Cell a,b,c ',[self.a, self.b, self.c], format='{:12.6f}')
        print_reals('Unit Cell alpha,beta,gamma',[self.alpha, self.beta, self.gamma], format='{:12.6f}')
        print_reals('lattice', self.lattice[0], format='{:12.6f}')
        print_reals('', self.lattice[1], format='{:12.6f}')
        print_reals('', self.lattice[2], format='{:12.6f}')
        print_strings('Element names',self.element_names)
        print_reals('Fractional coords',self.fractional_coordinates[0], format='{:12.6f}')
        for frac in self.fractional_coordinates[1:]:
            print_reals('',frac, format='{:12.6f}')
        print_reals('Cartesian coords',self.xyz_coordinates[0], format='{:12.6f}')
        for xyz in self.xyz_coordinates[1:]:
            print_reals('',xyz, format='{:12.6f}')
        if self.molecules:
            for molid,atoms in enumerate(self.molecules):
                mass, cm_xyz, cm_frac = self.calculateCentreOfMass(atom_list=atoms,units='all')
                molstring = 'Molecule '+str(molid)+':'
                print_ints('Atoms in '+molstring,atoms)
                print_reals('Mass of '+molstring,[ mass ], format='{:12.6f}')
                print_reals('Centre of Mass  (xyz) of '+molstring, cm_xyz, format='{:12.6f}')
                print_reals('Centre of Mass (frac) of '+molstring, cm_frac,format='{:12.6f}')

    def calculateCentreOfMass(self,atom_list=None, units='xyz'):
        # Calculate the centre of mass - if the atom list is given just use that
        # The centre of mass can be returned in units of 'xyz' space or 'abc' space
        # if units='all' a tuple of (mass,cm_xyz,cm_abc) is returned
        if atom_list == None:
            atom_list = range(self.nions)
        mass = 0.0
        cm = np.zeros(3)
        for atom_index in atom_list:
            mass += self.atomic_masses[atom_index]
            cm = cm + self.atomic_masses[atom_index] * self.xyz_coordinates[atom_index]
        cm_xyz = cm / mass
        if units == 'xyz':
            return cm_xyz
        elif units == 'mass':
            return mass
        elif units == 'abc':
            cm_fractional = self.convert_xyz_to_abc(cm_xyz)
            return cm_fractional
        else:
            cm_fractional = self.convert_xyz_to_abc(cm_xyz)
            return mass, cm_xyz, cm_fractional

    def set_atomic_masses(self, masses):
        self.atomic_masses = masses

    def convert_unitcell_to_abc(self):
        """Convert a unit cell to the equivalent a, b, c, alpha, beta, gamma designation"""
        a = np.sqrt(np.dot(self.lattice[0], self.lattice[0]))
        b = np.sqrt(np.dot(self.lattice[1], self.lattice[1]))
        c = np.sqrt(np.dot(self.lattice[2], self.lattice[2]))
        gamma = np.arccos(np.dot(self.lattice[0], self.lattice[1]) / (a*b))
        beta = np.arccos(np.dot(self.lattice[0], self.lattice[2]) / (a*c))
        alpha = np.arccos(np.dot(self.lattice[1], self.lattice[2]) / (b*c))
        return a, b, c, np.degrees(alpha), np.degrees(beta), np.degrees(gamma)

    def convert_abc_to_unitcell(self, a, b, c, alpha_degs, beta_degs, gamma_degs):
        """Convert a,b,c,alpha,beta,gamma to a unit cell"""
        # This is castep convention, need to check it works with vasp
        alpha = np.radians(alpha_degs)
        beta = np.radians(beta_degs)
        gamma = np.radians(gamma_degs)
        lattice = [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]]
        lattice[2] = [0.0, 0.0, c]
        lattice[1] = [0.0, b*np.sin(alpha), b*np.cos(alpha)]
        z = a * np.cos(beta)
        y = a * (np.cos(gamma) - np.cos(alpha)*np.cos(beta))/np.sin(alpha)
        x = a * np.sqrt(1.0 - np.cos(beta)**2 - (y/a)**2)
        lattice[0] = [x, y, z]
        self._calculate_reciprocal_lattice(lattice)
        return self.lattice

    def _calculate_reciprocal_lattice(self, lattice):
        """Calculate the reciprocal lattice"""
        self.lattice = np.array(lattice)
        self.a, self.b, self.c, self.alpha, self.beta, self.gamma = self.convert_unitcell_to_abc()
        self.volume = self.calculate_volume()
        self.inverse_lattice = np.linalg.inv(self.lattice)
        self.reciprocal_lattice = self.inverse_lattice

    def calculate_volume(self):
        """Calculate the volume"""
        volume = np.abs(np.dot(self.lattice[0], np.cross(self.lattice[1], self.lattice[2])))
        return volume

    def convert_xyz_to_abc(self, xyz):
        """Convert xyz coordinates to abc lattice coordinates"""
        xyz = np.array(xyz)
        abc = np.dot(xyz, self.reciprocal_lattice)
        return abc

    def convert_hkl_to_xyz(self, hkl):
        """Convert hkl coordinates to xyz coordinates"""
        hkl = np.array(hkl)
        xyz = np.dot(hkl, self.reciprocal_lattice.T)
        return xyz

    def convert_abc_to_xyz(self, abc):
        """Convert abc coordinates to xyz coordinates"""
        abc = np.array(abc)
        xyz = np.dot(abc, self.lattice)
        return xyz

    def convert_hkl_to_xyz2(self, hkl):
        """Convert hkl coordinates to xyz coordinates using a different method"""
        # Use the hkl miller indices to calculate the normal to a plane in cartesian space
        p = []
        if not hkl[0] == 0:
            p.append(np.array(self.lattice[0] / hkl[0]))
        if not hkl[1] == 0:
            p.append(np.array(self.lattice[1] / hkl[1]))
        if not hkl[2] == 0:
            p.append(np.array(self.lattice[2] / hkl[2]))
        if hkl[0] == 0:
            p.append(self.lattice[0] + p[0])
        if hkl[1] == 0:
            p.append(self.lattice[1] + p[0])
        if hkl[2] == 0:
            p.append(self.lattice[2] + p[0])
        p12 = p[1] - p[0]
        p13 = p[2] - p[0]
        normal = np.cross(p12, p13)
        norm = np.linalg.norm(normal)
        if norm < 1.0e-8:
            print("Error in unit cell, calculation of normal to miller index failed")
            exit(1)
        normal = normal / norm
        return normal

    def set_fractional_coordinates(self, coords):
        """Set the fractional coordinates and calculate the xyz coordinates"""
        self.fractional_coordinates = coords
        self.xyz_coordinates = self.convert_abc_to_xyz(coords)
        self.nions = len(coords)
        return

    def set_xyz_coordinates(self, coords):
        """Set the xyz coordinates and calculate the fractional coordinates"""
        self.xyz_coordinates = coords
        self.fractional_coordinates = self.convert_xyz_to_abc(coords)
        self.nions = len(coords)
        return

    def set_element_names(self, element_names):
        self.element_names = []
        for el in element_names:
          el = cleanup_symbol(el)
          self.element_names.append(el)
        return

    def set_atom_labels(self, atom_labels):
        self.atom_labels = []
        for el in atom_labels:
          self.atom_labels.append(el)
        return

    def find_symmetry(self):
        """Find the space group symmetry of the unit cell"""
        from spglib import get_spacegroup
        cell = ( self.lattice, self.fractional_coordinates, self.atomic_nos )
        self.spacegroup = get_spacegroup(cell, symmprec=1e-5)
        print("Symmetry space group is", self.spacegroup)

    def calculate_molecular_contents(self, scale, toler, covalent_radii):
        """Find whole molecules in the unit cell
           scale and toler are used to manipulate the covalent radii, thus the largest distance
           apart atoms i and j can be for a bond is
           scale*( radi + radj) + toler """
        # Calculate the contents of all the cells adjacent to the central cell
        adjacents = ( 0, -1, 1 )
        translations = [ (i, j, k) for i in adjacents for j in adjacents for k in adjacents ]
        fractional_supercell = []
        index_supercell = []
        for tr in translations:
            for l,a in enumerate(self.fractional_coordinates):
                i,j,k = tr
                index_supercell.append( l )
                new_position = [ (xyz1 + xyz2)  for xyz1, xyz2 in zip(a, tr) ]
                fractional_supercell.append( new_position )
                #jk print('New positions ',new_position,l,i,j,k)
        # Convert fractional supercell coordinates to xyz
        # xyz_supercell will be an np array
        xyz_supercell = np.empty_like(fractional_supercell)
        for i,abc in enumerate(fractional_supercell):
            xyz_supercell[i] = self.convert_abc_to_xyz(abc)
        # put the atoms into boxes of boxSize
        BoxAtoms = {}
        BoxNeighbours = {}
        boxSize = 0.0
        # calculate boxsize
        rmax = 0.0
        for el in self.element_names:
            #jk print("Element name",el)
            if covalent_radii[el] > rmax:
                rmax = covalent_radii[el]
        boxSize = 2.0*scale*rmax + 0.5 + toler
        #jk print('rmax = ',rmax)
        # Put atoms into boxes and store the box info in Atom_box_id
        Atom_box_id = []
        for i,xyz in enumerate(xyz_supercell):
            a = int( math.floor(xyz[0]/boxSize) )
            b = int( math.floor(xyz[1]/boxSize) )
            c = int( math.floor(xyz[2]/boxSize) )
            abc = (a,b,c)
            Atom_box_id.append(abc)
            try:
                BoxAtoms[abc].append(i)
            except:
                BoxAtoms[abc] = [i]
        # Calculate the neighbouring boxes for each occupied box
        for abc in BoxAtoms:
            #jk print('Box ',abc, BoxAtoms[abc])
            a,b,c = abc
            BoxNeighbours[abc] = []
            for i in [ -1, 0, 1]:
              for j in [ -1, 0, 1]:
                for k in [ -1, 0, 1]:
                  BoxNeighbours[abc].append( (a+i,b+j,c+k) )
        # end for abc in Box1

        # Calculate the bonding the supercell
        bondedToAtom = {}
        for i,xyzi in enumerate(xyz_supercell):
            #jk print('Calculating bonding to ',i,xyzi)
            bondedToAtom[i] = []
            # Find the element name for this atom in the supercell
            ip = index_supercell[i]
            i_el = self.element_names[ip]
            # lookup all of the boxes that might hold a bonding atom
            cell = Atom_box_id[i]
            for abc in BoxNeighbours[cell]:
                # lookup all the atoms that are in that cell
                try:
                    for j in BoxAtoms[abc]:
                        if j < i:
                            # Find the element name for this atom in the supercell
                            jp = index_supercell[j]
                            j_el = self.element_names[jp]
                            dist1 = scale*( covalent_radii[i_el] + covalent_radii[j_el]) + toler
                            dist2 = calculate_distance(xyzi,xyz_supercell[j])
                            if dist2 < dist1:
                                bondedToAtom[i].append(j)
                                bondedToAtom[j].append(i)
                                #jk print('new bond', i, j, i_el, j_el, xyzi, xyz_supercell[j], dist1, dist2)
                            # end if dist2 < dist1
                        # end if j < i
                    # end for j
                except KeyError:
                    pass
            # end for abc
        # end for i,a
        #
        # Now we have to find how many molecules we have in the cell
        # There must be at least one molecule in the cell and it must contain the first atom
        #
        # BelongsToMolecule[i] holds which molecule the ith atom belongs to
        belongsToMolecule = {}
        # molecules is a dictionary of molecules each entry is a is a list of atoms
        molecules = {}
        molID = -1
        # We stop when all the atoms in the original cell belong to a molecule
        remainingAtoms = [ atom for atom in range(self.nions) ]
        bonds = []
        while len(remainingAtoms) > 0:
            #jk print("Remaining atoms")
            #jk print(remainingAtoms)
            # create a new molecule from the first atom which has no molecule assigned to it
            molID += 1
            useAtom = remainingAtoms[0]
            belongsToMolecule[useAtom] = molID
            molecules[molID] = [ useAtom ]
            remainingAtoms.remove(useAtom)
            # Now using this new molecule with its only atom as a seed find any atoms connected to it
            # We need to continue searching for bonded atoms until there are no more to be found
            moreAtomsToBeFound = True
            while moreAtomsToBeFound:
                moreAtomsToBeFound = False
                for i in range(len(xyz_supercell)):
                    # has this atom been assigned a molecule yet?
                    if i in belongsToMolecule:
                        # atom i is already assigned to a molecule
                        useThisMolecule = belongsToMolecule[i]
                        #jk print("Using this molecule", useThisMolecule)
                        # Go through all the atoms bonded to i and add to the current molecule
                        for j in bondedToAtom[i]:
                            jx = index_supercell[j]
                            #jk print("atom j / jx is bonded to atom i",j,jx,i)
                            # The image of j in the original cell might not be available, and j might be bonded
                            if jx in remainingAtoms and not j in belongsToMolecule:
                                # if j was not already in a molecule then we have new information
                                moreAtomsToBeFound = True
                                molecules[useThisMolecule].append(j)
                                belongsToMolecule[j] = useThisMolecule
                                #jk print("Removing atom index(j) from remaining atoms",index_supercell[j])
                                remainingAtoms.remove(jx)
                            # The j'th atom could be already specified and we have a ring....
                            # We also need to make sure that we have unique bonds
                            if j in belongsToMolecule:
                                if i > j and (i,j) not in bonds:
                                    bonds.append( (i,j) )
                                elif i < j and (j,i) not in bonds:
                                    bonds.append( (j,i) )
                        # end for j
                    # end if i in
                # end for i
            # while moreAtomsToBeFound
        # until all the atoms belong to a molecule
        #jk print('Number of molecules', molID+1)
        self.centres_of_mass = []
        self.total_mass = 0.0
        for mol_index in molecules:
            #jk print('Molecule ',mol_index)
            #jk print('Atoms ',molecules[mol_index])
            for atom_index in molecules[mol_index]:
                index = index_supercell[atom_index]
                #jk print('New atom index, old index', atom_index, index, self.element_names[index])
            # Calculate centre of mass
            mass = 0.0
            cm = np.zeros(3)
            for atom_index in molecules[mol_index]:
                index = index_supercell[atom_index]
                mass += self.atomic_masses[index]
                cm = cm + self.atomic_masses[index] * xyz_supercell[atom_index]
            cm_xyz = cm / mass
            cm_fractional = self.convert_xyz_to_abc(cm_xyz)
            self.centres_of_mass.append( cm_fractional )
            #jk print('Mass', mass)
            #jk print('Centre of mass', cm_fractional)
        # Create a new unit cell with the atoms shifted so that whole molecules are ordered and within the cell
        new_molecules = []
        new_fractional = np.empty_like(self.fractional_coordinates)
        new_element_names = []
        new_index = 0
        new_masses = []
        old_order = []
        for mol_index in molecules:
            new_atom_index = []
            cm = self.centres_of_mass[mol_index]
            shift = np.array( [ 0.0, 0.0, 0.0] )
            if cm[0] < 0.0:
                shift += [ 1.0, 0.0, 0.0 ]
            elif cm[0] > 1.0:
                shift += [-1.0, 0.0, 0.0 ]
            if cm[1] < 0.0:
                shift += [ 0.0, 1.0, 0.0 ]
            elif cm[1] > 1.0:
                shift += [ 0.0,-1.0, 0.0 ]
            if cm[2] < 0.0:
                shift += [ 0.0, 0.0, 1.0 ]
            elif cm[2] > 1.0:
                shift += [ 0.0, 0.0,-1.0 ]
            for atom_index in molecules[mol_index]:
                old_index = index_supercell[atom_index]
                old_order.append(old_index)
                new_fractional[new_index] = shift + np.array(fractional_supercell[atom_index])
                new_element_names.append(self.element_names[old_index])
                new_masses.append(self.atomic_masses[old_index])
                new_atom_index.append(new_index)
                new_index += 1
            new_molecules.append(new_atom_index)
        # as well as being able to go from the new order and look up the old order
        # we need to be able to take the old order and look up what the new order is
        invert_old_order = np.zeros_like(old_order)
        for i,j in enumerate(old_order):
            invert_old_order[j] = i
        new_bonds = []
        for bond in bonds:
            i,j = bond
            ix = invert_old_order[index_supercell[i]]
            jx = invert_old_order[index_supercell[j]]
            new_bonds.append( (ix,jx) )
        new_unit_cell = UnitCell( self.a, self.b, self.c, self.alpha, self.beta, self.gamma )
        new_unit_cell.set_fractional_coordinates(new_fractional.tolist())
        new_unit_cell.set_element_names(new_element_names)
        new_unit_cell.set_atomic_masses(new_masses)
        new_unit_cell.set_molecules(new_molecules)
        new_unit_cell.set_bonds(new_bonds)
        return new_unit_cell, len(new_molecules), old_order

    def set_bonds(self, bonds):
        '''Define a list of bonds for the unit cell'''
        self.bonds = bonds
        return

    def set_molecules(self, molecules):
        '''Define a list of molecules, each molecule is a list of atom coordinates'''
        self.molecules = molecules
        return




#!/usr/bin/env python
"""Read in two csv files and check that the float contents are the same"""
from __future__ import print_function
import string
import os, sys
from os.path import splitext
from termcolor import colored
global separator
global threshold

def compare_lines( line1, line2, swapped):
  global separator
  global threshold
  store_error = False
  nerrors = 0
  percentage_error = 0.0
  max_percentage_error = 0.0
  keep_word1 = ''
  keep_word2 = ''
  index = 0
  len1 = len(line1.split(separator))
  len2 = len(line2.split(separator))
  if len1 == len2:
      if len1 > 4:
          word1 = line1.split(separator)[3]
          word2 = line2.split(separator)[3]
          if word1 == word2  and word1 == "processors":
              # Ignore this line 'cos the number of processors used depends on the machine
              return (store_error, nerrors, max_percentage_error, keep_word1, keep_word2)
          # end if word1 == word2
      # end if len1 > 4
  #end if len1 == len2
  for word1,word2 in zip(line1.split(separator),line2.split(separator)):
    index += 1
    if isfloat(word1):
        if not isfloat(word2):
            nerrors +=1
            store_error = True
        else:
            float1 = float(word1)
            float2 = float(word2)
            error = abs(float1 - float2)
            error2 = abs(abs(float1 - float2) -1.0)
            # This complicated test is there so that if the lines have been swapped and the first number
            # is an index number we ignore the index in the comparison
            # EG:  File1                                File2
            #      5 0.99999                            5 0.88888
            #      6 0.88888                            6 0.99999
            if not swapped or index > 1 or error2 > 1.0e-10:
                percentage_error = 100.0*error/(abs(float1)+abs(float2)+2.0)/2.0
                if percentage_error > threshold:
                    nerrors += 1
                    store_error = True
                    if max_percentage_error < percentage_error:
                        keep_word1 = word1
                        keep_word2 = word2
                        max_percentage_error = percentage_error
                # end if
             #end if error_swapped
        #end if not
    #end if isfloat
  # end for word1, word2
  return (store_error, nerrors, max_percentage_error, keep_word1, keep_word2)


def isfloat(value):
  try:
    float(value)
    return True
  except ValueError:
    return False


# Start processing the directories
if len(sys.argv) <= 1 :
    print('checkcsv file file2 [-sep blanks] [-thresh 3.0E-2]', file=sys.stderr)
    print('         Compare the two files for any significant changes', file=sys.stderr)
    print('         The default separator is "blanks" for txt/out files', file=sys.stderr)
    print('         and "," for csv files', file=sys.stderr)
    print('         Numbers f1 and f2 in each file are compared', file=sys.stderr)
    print('         an error is flagged if 2*abs(f1-f2)/(f1+f2+2) > threshold', file=sys.stderr)
    print('         The default value for threshold is 3.0E-2 ', file=sys.stderr)
    exit()

threshold = 3.0E-2
separator = None
tokens = sys.argv[1:]
ntokens = len(tokens)-1
itoken = -1
files = []
while itoken < ntokens:
    itoken += 1
    token = tokens[itoken]
    if token == '-sep':
        itoken +=1
        separator = tokens[itoken]
        if separator == 'blanks':
            separator = None
    elif token == '-thresh':
        itoken +=1
        threshold = float(tokens[itoken])
    else:
        files.append(tokens[itoken])
    # end if
# end while
file1 = files[0]
file2 = files[1]
file_name,extension = splitext(file1)
if extension == '.csv':
    separator = ','
elif extension == '.txt':
    separator = None
elif extension == '.out':
    separator = None
# Read the command line again
files = []
while itoken < ntokens:
    itoken += 1
    token = tokens[itoken]
    if token == '-sep':
        itoken +=1
        separator = tokens[itoken]
        if separator == 'blanks':
            separator = None
    elif token == '-thresh':
        itoken +=1
        threshold = float(tokens[itoken])
    else:
        files.append(tokens[itoken])
    # end if
# end while
fd1 = open(file1)
fd2 = open(file2)
#
line1_errors = []
line2_errors = []
line_numbers_with_error = []
nerrors = 0
store_error = False
max_percentage_error = 0.0
max_percentage_error2= 0.0
line_number = 0
compare_next_line = False
for line1,line2 in zip(fd1,fd2):
  line_number += 1
  if not compare_next_line:
      store_error, nerror, percentage_error, keep_word1c, keep_word2c = compare_lines(line1,line2,False)
      max_percentage_error = max(max_percentage_error,percentage_error)
      if store_error:
          # if there was an error let's see if two lines are swapped
          line1a = line1
          line2a = line2
          compare_next_line = True
  else:
      # There was an error on the previous line, see if lines are swapped
      store_error1, nerror1, percentage_error1, keep_word1a, keep_word2a = compare_lines(line1,line2a,True)
      store_error2, nerror2, percentage_error2, keep_word1b, keep_word2b = compare_lines(line2,line1a,True)
      compare_next_line = False
      keep_word1 = keep_word1a
      keep_word2 = keep_word2a
      keep_line_number = line_number
      if store_error1 or store_error2:
          # There is still an error on the swapped lines so we have an error on line1 and line1a
          nerrors += nerror
          if percentage_error > max_percentage_error:
              max_percentage_error = percentage_error
              keep_word1 = keep_word1c
              keep_word2 = keep_word2c
              keep_line_number = line_number - 1
          store_error, nerror, percentage_error, keep_word1c, keep_word2c = compare_lines(line1,line2,False)
          nerrors += nerror
          if percentage_error > max_percentage_error:
              max_percentage_error = percentage_error
              keep_word1 = keep_word1c
              keep_word2 = keep_word2c
              keep_line_number = line_number
      else:
          # The swapped lines are the same so there is no error
          store_error = False
      # endif
   # end if not
#end for line1, line2
if nerrors > 0:
  print('  '+colored('ERRORS:','red')+"({}) LARGEST ON LINE {} OF {}({}) and {}({}) -- max %error={}".format(nerrors, keep_line_number, file1,keep_word1,file2,keep_word2,max_percentage_error))
else:
  print('  '+colored('OK:','blue')+" {} = {} -- max %error={}" .format(file1,file2,max_percentage_error))



#!/usr/bin/env python
"""Read the contents of a directory containing DFT output and create a csv style file of information"""
from __future__ import print_function
import string
import re
import numpy as np
import os, sys
from PDielec.Constants import amu, PI, avogadro_si, wavenumber, angstrom, isotope_masses, average_masses
from PDielec.VaspOutputReader import VaspOutputReader
from PDielec.CastepOutputReader import CastepOutputReader
from PDielec.GulpOutputReader import GulpOutputReader
from PDielec.CrystalOutputReader import CrystalOutputReader
from PDielec.AbinitOutputReader import AbinitOutputReader
from PDielec.QEOutputReader import QEOutputReader
from PDielec.PhonopyOutputReader import PhonopyOutputReader
from multiprocessing import Pool, cpu_count
import PDielec.Calculator as Calculator

def set_affinity_on_worker():
    """When a new worker process is created, the affinity is set to all CPUs"""
    #JK print("I'm the process %d, setting affinity to all CPUs." % os.getpid())
    #JK for the time being this is simply commented out, but might be useful at some point
    #os.system("taskset -p 0xff %d > /dev/null" % os.getpid())

def read_a_file( calling_parameters):
    name, eckart, neutral, mass_definition, mass_dictionary, global_no_calculation, program, hessian_symmetrisation, qmprogram, debug = calling_parameters
    fulldirname = name
    head,tail = os.path.split(fulldirname)
    root,ext = os.path.splitext(tail)
    if program == "castep":
        names = [ name ]
        reader = CastepOutputReader( names )
    elif program == "vasp":
        name1 = os.path.join(head,'OUTCAR')
        name2 = os.path.join(head,'KPOINTS')
        names = [ name1, name2 ]
        reader = VaspOutputReader( names )
    elif program == "gulp":
        names = [ name ]
        reader = GulpOutputReader( names )
    elif program == "crystal":
        names = [ name ]
        reader = CrystalOutputReader( names )
    elif program == "abinit":
        names = [ name ]
        reader = AbinitOutputReader( names )
    elif program == "qe":
        tail1 = root+'.dynG'
        tail2 = root+'.log'
        tail3 = root+'.out'
        name1 = os.path.join(head,tail2)
        name2 = os.path.join(head,tail3)
        name3 = os.path.join(head,tail)
        # We want the dynG entry last rounding causes problems otherwise
        name4 = os.path.join(head,tail1)
        names = []
        for n in [ name1, name2, name3, name4 ]:
            if os.path.isfile(n):
                if not n in names:
                    names.append(n)
        reader = QEOutputReader( names )
    elif program == "phonopy":
        # The order is important
        pname1 = os.path.join(head,'qpoints.yaml')
        pname2 = os.path.join(head,'phonopy.yaml')
        # Only works for VASP at the moment
        vname1 = os.path.join(head,'OUTCAR')
        vname2 = os.path.join(head,'KPOINTS')
        pnames = [ pname1, pname2 ]
        vnames = [ vname1, vname2 ]
        pnames.extend(vnames)
        names = pnames
        # Which QM program was used by PHONOPY?
        if qmprogram == "castep":
            print("Error in qmreader",qmprogram)
            exit()
            qmreader = CastepOutputReader(names)
        elif qmprogram == "vasp":
            qmreader = VaspOutputReader(vnames)
        elif qmprogram == "gulp":
            print("Error in qmreader",qmprogram)
            exit()
            qmreader = GulpOutputReader(names)
        elif qmprogram == "crystal":
            print("Error in qmreader",qmprogram)
            exit()
            qmreader = CrystalOutputReader(names)
        elif qmprogram == "abinit":
            print("Error in qmreader",qmprogram)
            exit()
            qmreader = AbinitOutputReader(names)
        elif qmprogram == "qe":
            print("Error in qmreader",qmprogram)
            exit()
            qmreader = QEOutputReader(names)
        # The QM reader is used to get info about the QM calculation
        reader = PhonopyOutputReader(pnames,qmreader)
    else:
        print('Program name not recognized',program,file=sys.stderr)
        exit()

    print('  Analysing ',names, file=sys.stderr)
    # The order that the settings are applied is important
    # Symmetry is applied before the file is read, it is an integral part of the hessian
    # Eckart and neutral are applied after the file has been read, this way the original frequencies are those before any calculations
    reader.debug = debug
    reader.hessian_symmetrisation = hessian_symmetrisation
    reader.read_output()
    frequencies_cm1 = reader.frequencies
    frequencies = np.array(frequencies_cm1)
    no_calculation = global_no_calculation
    if len(frequencies) < 3:
        no_calculation = True
    mode_list = []
    ignore_modes = []
    sigmas = []
    if not no_calculation:
        # apply the eckart conditions before we change the masses
        reader.eckart = eckart
        # Get the born charges
        if neutral:
            reader.neutralise_born_charges()
        # What mass definition are we using?
        if not mass_definition == "program" or mass_dictionary:
            if mass_definition == "average":
                reader.change_masses(average_masses, mass_dictionary)
            elif mass_definition == "isotope":
                reader.change_masses(isotope_masses, mass_dictionary)
        born_charges = np.array(reader.born_charges)
        # Calculate the mass weighted normal modes.  This just forces projection
        mass_weighted_normal_modes = reader.calculate_mass_weighted_normal_modes()
        # The masses might have changed in the calculation of the mass weighted normal modes
        masses = np.array(reader.masses) * amu
        normal_modes = Calculator.normal_modes(masses, mass_weighted_normal_modes)
        # from the normal modes and the born charges calculate the oscillator strengths of each mode
        oscillator_strengths = Calculator.oscillator_strengths(normal_modes, born_charges)
        modified_frequencies_cm1 = reader.frequencies
        modified_frequencies_cm1.sort()
        modified_frequencies = np.array(modified_frequencies_cm1)
        # if the frequency is less than 5 cm-1 assume that the oscillator strength is zero
        for imode,f in enumerate(modified_frequencies_cm1):
            mode_list.append(imode)
            sigmas.append(5.0)
            if f < 5.0:
                oscillator_strengths[imode]=np.zeros((3,3))
        # calculate the intensities from the trace of the oscillator strengths
        intensities = Calculator.infrared_intensities(oscillator_strengths)
        # Calculate eps0
        volume = reader.volume*angstrom*angstrom*angstrom
        #
        # Calculate degenerate lists
        #
        degeneracy_threshold = 1.0E-8
        threshold_intensity = 1.0E-6
        threshold_frequency = 5.0
        degenerate_lists = {}
        mmax = len(modified_frequencies_cm1)
        for m1 in range(len(modified_frequencies_cm1)):
            degenerate_lists[m1] = []
        for m1 in range(len(modified_frequencies_cm1)):
            f1 = modified_frequencies[m1]
            for m2 in range(m1+1,min(mmax,m1+3)):
                f2 = modified_frequencies_cm1[m2]
                if abs(f2-f1) < degeneracy_threshold:
                    degenerate_lists[m1].append(m2)
                    degenerate_lists[m2].append(m1)
        #
        # Only modes with non-zero oscillator strengths contribute to the dielectric
        # so calculate those modes which we can safely ignore and store them in ignore_modes
        #
        if len(ignore_modes) == 0:
            for mode, intensity in enumerate(intensities):
                # ignore modes with a low oscillator strength
                if intensity < threshold_intensity:
                    # If any of its degenerate modes have intensity then we shouldn't ignore it
                    ignore = True
                    for m in degenerate_lists[mode]:
                        if intensities[m] > threshold_intensity:
                            ignore = False
                    if ignore:
                        ignore_modes.append(mode)
                # ignore modes with low real frequency
                elif np.real(modified_frequencies_cm1[mode])/wavenumber < threshold_frequency:
                    ignore_modes.append(mode)
                # ignore modes with imaginary frequency
                elif abs(np.imag(modified_frequencies_cm1[mode]))/wavenumber > 1.0e-6:
                    ignore_modes.append(mode)
                # end if intensity
            # end for
        # end if len()
        # Remove any unwanted modes
        ignore_modes = list(set(ignore_modes))
        if len(ignore_modes) > 0:
            for mode in ignore_modes:
                if mode in mode_list:
                    mode_list.remove(mode)
            # end loop over modes to be ignored
        # end of if ignore_modes
        ionicv = Calculator.dielectric_contribution(0.0, mode_list, modified_frequencies*wavenumber, sigmas, oscillator_strengths, volume)
    # absorption units here are L/mole/cm-1
    # Continue reading any data from the output file
    frequencies_cm1.sort()
    unitCell = reader.unit_cells[-1]
    a,b,c,alpha,beta,gamma = unitCell.convert_unitcell_to_abc()
    if not no_calculation:
        eps0 = np.real(ionicv)
    else:
        eps0   = np.array(reader.zerof_static_dielectric)
    epsinf = np.array(reader.zerof_optical_dielectric)
    eps0_xx = str(eps0[0,0])
    eps0_yy = str(eps0[1,1])
    eps0_zz = str(eps0[2,2])
    eps0_xy = str(( eps0[0,1] + eps0[1,0] ) /2.0)
    eps0_xz = str(( eps0[0,2] + eps0[2,0] ) /2.0)
    eps0_yz = str(( eps0[1,2] + eps0[2,1] ) /2.0)
    epsinf_xx = str(epsinf[0,0])
    epsinf_yy = str(epsinf[1,1])
    epsinf_zz = str(epsinf[2,2])
    epsinf_xy = str(( epsinf[0,1] + epsinf[1,0] ) /2.0)
    epsinf_xz = str(( epsinf[0,2] + epsinf[2,0] ) /2.0)
    epsinf_yz = str(( epsinf[1,2] + epsinf[2,1] ) /2.0)
    volume = str(reader.volume)
    carray = np.array(reader.elastic_constants)
    c11 = str(carray[0,0])
    c22 = str(carray[1,1])
    c33 = str(carray[2,2])
    c44 = str(carray[3,3])
    c55 = str(carray[4,4])
    c66 = str(carray[5,5])
    c12 = str(carray[0,1])
    c13 = str(carray[0,2])
    c23 = str(carray[1,2])
    # Assemble the output line for the unprojected frequencies and all the other data
    header = fulldirname + ',Read from program'
    string = ''
    string = string + ',' + str(reader.electrons)
    string = string + ',' + str(reader.magnetization)
    string = string + ',' + str(reader.kpoints)
    string = string + ',' + str(reader.kpoint_grid[0])
    string = string + ',' + str(reader.kpoint_grid[1])
    string = string + ',' + str(reader.kpoint_grid[2])
    string = string + ',' + str(reader.energy_cutoff)
    string = string + ',' + str(reader.final_free_energy)
    string = string + ',' + str(reader.final_energy_without_entropy)
    string = string + ',' + str(reader.pressure)
    string = string + ',' + str(a) + ',' + str(b) + ',' + str(c) + ',' + str(alpha) + ',' + str(beta) + ',' + str(gamma) + ',' + volume
    string = string + ',' + eps0_xx + ',' + eps0_yy + ',' + eps0_zz + ',' + eps0_xy + ',' + eps0_xz + ',' + eps0_yz
    string = string + ',' + epsinf_xx + ',' + epsinf_yy + ',' + epsinf_zz + ',' + epsinf_xy + ',' + epsinf_xz + ',' + epsinf_yz
    string = string + ',' + c11 + ',' + c22 + ',' + c33 + ',' + c44 + ',' + c55 + ',' + c66
    string = string + ',' + c12 + ',' + c13 + ',' + c23
    common_output=string
    string = header+common_output
    for f in frequencies_cm1:
        string = string + ',' + str(f)
    results_string = [ string ]
    # Assemble the next line if any of eckart/neutral/symm have been used
    option_string = ''
    if not no_calculation:
        option_string = 'Calculated frequencies (cm-1)'
        if eckart :
            option_string = option_string+' eckart'
        if neutral :
            option_string = option_string+' neutral'
        option_string = option_string+' hessian_symmetrisation='+hessian_symmetrisation
        option_string = option_string+' mass_definition='+mass_definition
        header = fulldirname+','+option_string
        string = header + common_output
        for f in modified_frequencies_cm1:
            string = string + ',' + str(f)
        results_string.append(string)
        option_string = 'Calculated Intensities (Debye2/Angs2/amu)'
        header = fulldirname+','+option_string
        string = header + common_output
        for f in intensities:
            string = string + ',' + str(f)
        results_string.append(string)
        option_string = 'Calculated Integrated Molar Absorption (L/mole/cm/cm)'
        header = fulldirname+','+option_string
        string = header + common_output
        for f in intensities:
            string = string + ',' + str(f*4225.6)
        results_string.append(string)
    # End if not no_calculation
    return name,results_string

def main():
    # Start processing the directories
    if len(sys.argv) <= 1 :
        print('preader -program program [-eckart] [-neutral] [-nocalculation] [-hessian symm] [-masses average] filenames .....', file=sys.stderr)
        print('  \"program\" must be one of \"abinit\", \"castep\", \"crystal\", \"gulp\"       ', file=sys.stderr)
        print('           \"phonopy\", \"qe\", \"vasp\"                                         ', file=sys.stderr)
        print('           If phonopy is used it must be followed by the QM package              ', file=sys.stderr)
        print('  -masses [average|isotope|program]  chooses the atomic mass definition average  ', file=sys.stderr)
        print('          The default is \"average\"                                             ', file=sys.stderr)
        print('  -mass  element mass                                                            ', file=sys.stderr)
        print('          Change the mass of the element specified                               ', file=sys.stderr)
        print('  -eckart projects out the translational components. Default is no projection    ', file=sys.stderr)
        print('  -neutral imposes neutrality on the Born charges of the molecule.               ', file=sys.stderr)
        print('           Default is no neutrality is imposed                                   ', file=sys.stderr)
        print('  -hessian [symm|crystal] defines the symmetrisation of the hessian              ', file=sys.stderr)
        print('           \"crystal\" imposes Crystal14 symmetrisation                          ', file=sys.stderr)
        print('           \"symm\" symmetrises by averaging the hessian with its transpose      ', file=sys.stderr)
        print('           \"symm\" is the default                                               ', file=sys.stderr)
        print('  -nocalculation requests no calculations are performed                          ', file=sys.stderr)
        print('           A single line is output with results obtained by reading the output   ', file=sys.stderr)
        print('           any of -mass -masses -eckart -neutral or -crystal are ignored         ', file=sys.stderr)
        print('  -debug   to switch on more debug information                                   ', file=sys.stderr)
        exit()

    observables=False
    files = []
    eckart = False
    neutral = False
    hessian_symmetrisation = "symm"
    mass_definition = "average"
    mass_dictionary = {}
    global_no_calculation = False
    tokens = sys.argv[1:]
    ntokens = len(tokens)-1
    itoken = -1
    program = ''
    qmprogram = ''
    debug = False
    while itoken < ntokens:
        itoken += 1
        token = tokens[itoken]
        if token == "-observables":
            observables = True
        elif token == "-debug":
            debug = True
        elif token == "-eckart":
            eckart = True
        elif token == "-neutral":
            neutral = True
        elif token == "-hessian":
            itoken += 1
            hessian_symmetrisation = tokens[itoken]
            if (not hessian_symmetrisation == "symm") and ( not hessian_symmetrisation == "crystal" ):
                print('-hessian must be followed by \"symm\" or \"crystal\"',file=sys.stderr)
                exit()
        elif token == "-masses":
            itoken += 1
            token = tokens[itoken]
            if token[0] == 'a':
                mass_definition = 'average'
            elif token[0] == 'i':
                mass_definition = 'isotope'
            elif token[0] == 'p':
                mass_definition = 'program'
            else:
                print('-masses must be followed by \"average\" \"isotope\" or \"program\"',file=sys.stderr)
                exit()
        elif token == "-mass":
            itoken += 1
            element = tokens[itoken]
            itoken += 1
            mass = float(tokens[itoken])
            mass_dictionary[element] = mass
        elif token == "-nocalculation":
            global_no_calculation = True
        elif token == "-program":
            itoken += 1
            program = tokens[itoken]
            if program == 'phonopy':
                itoken += 1
                qmprogram = tokens[itoken]
        else:
            files.append(token)

    if len(program) < 1:
        print('Please use -program to define the package used to generate the output files',file=sys.stderr)
        exit()

    if not program in ['abinit','castep','crystal','gulp','qe','vasp','phonopy']:
        print('Program is not recognised: ',program,file=sys.stderr)
        exit()

    if program == 'phonopy':
        if not qmprogram in ['abinit','castep','crystal','gulp','qe','vasp']:
            print('Phonopy QM program is not recognised: ',qmprogram,file=sys.stderr)
            exit()
        print('  QM program used by Phonopy is: ',qmprogram,file=sys.stderr)

    print('  Program is ',program,file=sys.stderr)

    for f in files:
        if not os.path.isfile(f):
            print('Error file requested for analysis does not exist',f,file=sys.stderr)
            exit()

    if global_no_calculation:
        print('  No calculations has been requested',file=sys.stderr)
        print('  The frequencies (if present) will come from the QM/MM program    ',file=sys.stderr)
        print('  No hessian symmetrisation be performed',file=sys.stderr)
        if eckart:
            print('  The -eckart flag will be ignored',file=sys.stderr)
            eckart = False
        if neutral:
            print('  The -neutral flag will be ignored',file=sys.stderr)
            neutral = False
        if not mass_definition == "program":
            print('  The average and isotope mass definitions will be ignored',file=sys.stderr)
            mass_definition = "program"
        if mass_dictionary:
            print('  The masses specified by the -mass directive will be ignored',file=sys.stderr)
            mass_dictionary = {}
    else:
        print('  Eckart is ',eckart,file=sys.stderr)
        print('  Neutral is ',neutral,file=sys.stderr)
        print('  Mass definition is ',mass_definition,file=sys.stderr)
        print('  Hessian symmetrisation is ',hessian_symmetrisation,file=sys.stderr)
    #
    # Create a pool of processors to handle reading the files
    #
    p = Pool(initializer=set_affinity_on_worker)
    # Create a tuple list of calling parameters
    calling_parameters = []
    files.sort()
    for name in files:
        calling_parameters.append( (name, eckart, neutral, mass_definition, mass_dictionary, global_no_calculation, program, hessian_symmetrisation, qmprogram, debug) )
    # Calculate the results in parallel
    results_map_object = p.map_async(read_a_file,calling_parameters)
    results_map_object.wait()
    results = results_map_object.get()
    # Convert the results into a dictionary
    results_dictionary = {}
    for name,strings in results:
        results_dictionary[name] = strings
    # Print out the results
    print('directory,information,electrons,magnetization,kpnts,kpnt_1,kpnt_2,kpnt_3,energy_cutoff_eV,final_free_energy_eV,final_energy_without_entropy_eV,pressure_GPa,a_A,b_A,c_A,alpha,beta,gamma,volume,eps0_xx,eps0_yy,eps0_zz,eps0_xy,eps0_xz,eps0_yz,epsinf_xx,epsinf_yy,epsinf_zz,epsinf_xy,epsinf_xz,epsinf_yz, c11_gpa, c22_gpa, c33_gpa, c44_gpa, c55_gpa, c66_gpa, c12_gpa,  c13_gpa, c23_gpa,f1,f2,f3,f4,f5,f6....')
    for name in files:
        for string in results_dictionary[name]:
            print(string)
    #
    # Process the observables switch
    # ( for the time being this is commented out )
    exit()
    for name in files :
        if observables:
            filename = directory+'.observables'
            if directory == '.' :
                filename = 'dot.observables'
            filename = filename.replace("/","_")
            print('filename', filename, file=sys.stderr)
            fd = open(filename,'w')
            print('observables', file=fd)
            if eps0[0,0] != 0.0 :
                print('sdlc', file=fd)
                print('1 1 ', eps0[0,0], file=fd)
                print('sdlc', file=fd)
                print('2 2 ', eps0[1,1], file=fd)
                print('sdlc', file=fd)
                print('3 3 ', eps0[2,2], file=fd)
                print('hfdlc', file=fd)
                print('1 1 ', epsinf[0,0], file=fd)
                print('hfdlc', file=fd)
                print('2 2 ', epsinf[1,1], file=fd)
                print('hfdlc', file=fd)
                print('3 3 ', epsinf[2,2], file=fd)
            if carray[0,0] != 0.0:
                print('elastic', file=fd)
                print('1 1 ', carray[0,0], file=fd)
                print('elastic', file=fd)
                print('2 2 ', carray[1,1], file=fd)
                print('elastic', file=fd)
                print('3 3 ', carray[2,2], file=fd)
                print('elastic', file=fd)
                print('4 4 ', carray[3,3], file=fd)
                print('elastic', file=fd)
                print('5 5 ', carray[4,4], file=fd)
                print('elastic', file=fd)
                print('6 6 ', carray[5,5], file=fd)
                print('elastic', file=fd)
                print('1 2 ', carray[0,1], file=fd)
                print('elastic', file=fd)
                print('1 3 ', carray[0,2], file=fd)
                print('elastic', file=fd)
                print('2 3 ', carray[1,2], file=fd)
            nfreq = len(frequencies_cm1)
            if nfreq > 4 :
                for n in range(nfreq):
                    nlast = n
                    if frequencies_cm1[n].real > 1.0E-6:
                        break
                    # end of if
                # end of for
                if frequencies_cm1[nlast].real > 1.0E-8 :
                    n = nlast
                    print('frequency', nfreq - nlast, file=fd)
                    for f in frequencies_cm1[nlast:] :
                        n = n + 1
                        print(n, f, file=fd)
            print('end', file=fd)
            fd.close()
    # End of for loop over files
# end of def main

if __name__ == "__main__":
    main(sys)
